package org.processmining.featuremerge.plugin;

import java.util.HashMap;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.factory.XFactory;
import org.deckfour.xes.factory.XFactoryRegistry;
import org.deckfour.xes.model.XAttribute;
import org.deckfour.xes.model.XAttributeContinuous;
import org.deckfour.xes.model.XAttributeDiscrete;
import org.deckfour.xes.model.XAttributeMap;
import org.deckfour.xes.model.XEvent;
import org.deckfour.xes.model.XLog;
import org.deckfour.xes.model.XTrace;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.featuremerge.confs.Configuration;
import org.processmining.featuremerge.dialogs.AttributesSelectorDialog;
import org.processmining.featuremerge.dialogs.NewAttributeDefinition;
import org.processmining.featuremerge.dialogs.ThresholdSelectorDialog;
import org.processmining.featuremerge.models.IndicatorValuesMatrix;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginCategory;
import org.processmining.framework.plugin.annotations.PluginLevel;
import org.processmining.framework.plugin.events.Logger.MessageLevel;
import org.processmining.kpiprocessenhancementworkflow.models.LiteralIndicatorValuesArray;

public class FeatureMergePlugin {
	
	@Plugin(
			name = "Definition of a discrete attribute from numeric indicators",
			level=PluginLevel.Regular,
			parameterLabels = { "Log" }, 
			returnLabels = {"Log enriched with the new discrete attribute" }, 
			returnTypes = {XLog.class}, 
			userAccessible = true,
			categories=PluginCategory.Filtering,
			help = "Produces a XLog with a new discrete attribute created by defining constraints, at trace level, over existing attributes. The attribute will be attached to the last event of every trace and it will be set to the number of defined constraints satisfied by the trace."
			)
	
	@UITopiaVariant(
			affiliation = "Marche Polytechnic University", 
			author = "Emanuele Storti", 
			email = "e.storti@univpm.it"
			)

	public XLog combineFeatures(UIPluginContext context, XLog inputLog) {
		// this data structure stores information about indicators 
		HashMap<String,HashMap<String,Object>> KPIs = new HashMap<String,HashMap<String,Object>>();
		XLog log = (XLog) inputLog.clone();
		
		context.getProgress().setMinimum(0);
        context.getProgress().setMaximum(log.size());
        context.getProgress().setCaption("Analysing attributes in traces");
        context.getProgress().setIndeterminate(false);
        context.getProgress().setValue(0);
        // update user log
		context.log("Analysing attributes in traces",MessageLevel.NORMAL);
		
		IndicatorValuesMatrix matrix=new IndicatorValuesMatrix();
		LiteralIndicatorValuesArray literalArray=new LiteralIndicatorValuesArray();
		
		//go through the log to retrieve the list of all indicators
		for (XTrace trace : log) {
			//get all events
			//for (XEvent event : trace) {
			
			//get only the last event for every trace
			XEvent event=trace.get(trace.size()-1);
			XAttributeMap map = event.getAttributes();
			// for each attribute that is either a discrete or a continuous XAttribute, store it to the data structure
			for (String kpi : map.keySet()) {
					
				XAttribute a=map.get(kpi);
				
				if(a instanceof XAttributeDiscrete || a instanceof XAttributeContinuous) {
					double value=Double.parseDouble(a.toString());
				
					// if the KPI is not in the data structure, initialize it
					if(!matrix.isAvailable(kpi)) {
						matrix.initializeIndicatorRow(kpi);
					}
					//add the value to the matrix
					matrix.addValueToRow(kpi, value);
				}
				else if(a instanceof XAttributeDiscrete) { // the value cannot be cast to a double (i.e., it is not numeric)
					String value=a.toString();
					literalArray.addValueToRow(kpi,value);
				}
			}

    		// and increment the progress bar
    		context.getProgress().inc();
		}
	
		Configuration conf=new Configuration();
		conf.setMatrix(matrix);
		conf.setLiteralKPIs(literalArray);
		
		log=manageInteraction(context, conf, log, 1);
		return log;
	}
		
	private XLog manageInteraction(UIPluginContext context,Configuration conf, XLog log, int step) {
		XLog outputLog=null;
		if(step==1) {
			context.log("Evaluating statistics on attributes",MessageLevel.NORMAL);
			AttributesSelectorDialog dialog = new AttributesSelectorDialog(context,conf);
			// Show the dialog. User can now change the configuration.
		    InteractionResult result = context.showWizard("Attribute selection", true, false, dialog);
		    if (result == InteractionResult.NEXT) {
		    	String[] selected=dialog.getSelectedRows();
		    	if(selected.length>0) {
		    		conf.setSelected(selected);
		    		outputLog=manageInteraction(context,conf,log,2);}
		    	else outputLog=manageInteraction(context,conf,log,1);
		    }
		    else if(result == InteractionResult.CANCEL)
	    		return log;
		}
		else if(step==2) {
			ThresholdSelectorDialog dialog2=new ThresholdSelectorDialog(context,conf);
	    	InteractionResult result2 = context.showWizard("Threshold selection", false, false, dialog2);
	    	if(result2 == InteractionResult.PREV) {
	    		conf.setSelected(null);
	    		outputLog=manageInteraction(context,conf,log,1);
	    	}
	    	else if(result2 == InteractionResult.NEXT) {
	    		conf.setThreshods(dialog2.getThresholds());
	    		 outputLog=manageInteraction(context,conf,log,3);
	    	}
	    	else if(result2 == InteractionResult.CANCEL)
	    		return log;
		}
		else if(step==3) {
			NewAttributeDefinition dialog3=new NewAttributeDefinition(context,conf);
	    	InteractionResult result3 = context.showWizard("Name definition", false, true, dialog3);
	    	if(result3 == InteractionResult.PREV) {
	    		conf.setThreshods(null);
	    		 outputLog=manageInteraction(context,conf,log,2);
	    	}
	    	else if(result3 == InteractionResult.FINISHED) {
	    		String attribute=dialog3.getNewAttribute();
	    		if(attribute.length()==0)
	    			outputLog=manageInteraction(context,conf,log,3);
	    		else {
	    		//add to every trace the new attribute
	    		int i=0;
	    		context.getProgress().setMinimum(0);
                context.getProgress().setMaximum(log.size());
                context.getProgress().setCaption("Adding the new attribute "+attribute+" to traces");
                context.getProgress().setIndeterminate(false);
                context.getProgress().setValue(0);
	    		
                for(XTrace trace:log) {
	                XEvent lastEvent=trace.get(trace.size()-1);
	                XFactory factory = XFactoryRegistry.instance().currentDefault();
	                XAttributeDiscrete xAttr=factory.createAttributeDiscrete(attribute, 0, null);
	                
	                String value;
	                //check if the trace satisfy constraints
	                long checkNum=DistributionCalculator.checkContraintSatisfaction(conf, lastEvent);
	                
	        		xAttr.setValue(checkNum);
	        		lastEvent.getAttributes().put(attribute, xAttr);
	        		context.getProgress().inc();   
	        		i++;
	    		}
	    		
	    		// Do the heavy lifting.
	    		outputLog=log;		
	    		}}
	    	else if(result3 == InteractionResult.CANCEL)
	    		return log;
	}
		return outputLog;
	}
		
}

/**
 for (XTrace trace : log) {
			//get all events
			//for (XEvent event : trace) {
			
			//get only the last event for every trace
			XEvent event=trace.get(trace.size()-1);
				XAttributeMap map = event.getAttributes();
				// for each attribute that is either a discrete or a continuous XAttribute, store it to the data structure
				for (String kpi : map.keySet()) {
					
					XAttribute a=map.get(kpi);
					
					try {
						double value=Double.parseDouble(a.toString());
						
						// if the KPI is not in the data structure
						if(!KPIs.containsKey(kpi)) {
							
							//setup an entry for it
							HashMap<String,Object> hm=new HashMap<String,Object>();
							//hm.put("type", a.getClass().toString());
							hm.put("min",value);
							hm.put("max",value);
							hm.put("avg",value);
							KPIs.put(kpi,hm);
						}
						//else if KPI is in the data structure
						else {
							//update min/max if needed
							HashMap<String,Object> hm=KPIs.get(kpi);
							if(Double.compare(Double.parseDouble(a.toString()),(double)hm.get("min"))<0)
								hm.replace("min",Double.parseDouble(a.toString()));
							if(Double.compare(Double.parseDouble(a.toString()),(double)hm.get("max"))>0)
								hm.replace("max",Double.parseDouble(a.toString()));
							//update the avg
							hm.replace("avg", Double.parseDouble((hm.get("avg").toString()))+value);
						}
					}
					
			
						
					
					catch(Exception e) { // the value cannot be cast to a double (i.e., it is not numeric)
						continue;
					}
				}
**/
